@@ -36,11 +36,18 @@ Floating point printing cannot be made to work like the XS version.
 The XS version tests for NV or IV directly, but it is next to
 impossible to get this information from Perl without XS.
 
+=head1 TESTING
+
+To test this module, do
+
+    make
+    JSONCreatePP=1 make test
+
 =cut
 
 package JSON::Create::PP;
 use parent Exporter;
-our @EXPORT_OK = qw/create_json/;
+our @EXPORT_OK = qw/create_json create_json_strict json_escape/;
 our %EXPORT_TAGS = (all => \@EXPORT_OK);
 use warnings;
 use strict;
@@ -49,6 +56,7 @@ use Carp qw/croak carp confess cluck/;
 use Scalar::Util qw/looks_like_number blessed reftype/;
 use Unicode::UTF8 qw/decode_utf8 valid_utf8/;
 use B;
+our $VERSION = '0.23';
 
 # http://stackoverflow.com/questions/1185822/how-do-i-create-or-test-for-nan-or-infinity-in-perl#1185828
 
@@ -67,9 +75,9 @@ sub isnan {
 sub isfloat
 {
     my ($num) = @_;
-#  print "Looking at $num...";
+
     if ($num != int ($num)) {
-#	print "it's obviously a float.\n";
+	# It's clearly a floating point number
 	return 1;
     }
 
@@ -82,16 +90,7 @@ sub isfloat
     # Perl programmer.
 
     my $r = B::svref_2object (\$num);
-#    print "$r\n";
-    my $isfloat = $r->isa("B::NV") || $r->isa("B::PVNV");# && ! $r->isa ('B::IV');
-    # if ($isfloat) {
-    # 	print "it's a float";
-    # }
-    # else {
-    # 	print "it's not a float";
-    # }
-    # print "\n";
-
+    my $isfloat = $r->isa("B::NV") || $r->isa("B::PVNV");
     return $isfloat;
 }
 
@@ -102,11 +101,14 @@ sub isbool
     my ($ref) = @_;
     my $poo = B::svref_2object ($ref);
     if (ref $poo eq 'B::SPECIAL') {
-#	if ($B::specialsv_name[$$poo] eq '&PL_sv_yes') {
+	# Leave the following commented-out code as reference for what
+	# the magic numbers mean.
+
+	# if ($B::specialsv_name[$$poo] eq '&PL_sv_yes') {
 	if ($$poo == 2) {
 	    return 'true';
 	}
-#	elsif ($B::specialsv_name[$$poo] eq '&PL_sv_no') {
+	# elsif ($B::specialsv_name[$$poo] eq '&PL_sv_no') {
 	elsif ($$poo == 3) {
 	    return 'false';
 	}
@@ -114,6 +116,19 @@ sub isbool
     return undef;
 }
 
+sub json_escape
+{
+    my ($input) = @_;
+    $input =~ s/("|\\)/\\$1/g;
+    $input =~ s/\x08/\\b/g;
+    $input =~ s/\f/\\f/g;
+    $input =~ s/\n/\\n/g;
+    $input =~ s/\r/\\r/g;
+    $input =~ s/\t/\\t/g;
+    $input =~ s/([\x00-\x1f])/sprintf ("\\u%04x", ord ($1))/ge;
+    return $input;
+}
+
 sub escape_all_unicode
 {
     my ($jc, $input) = @_;
@@ -132,8 +147,10 @@ sub escape_all_unicode
 sub stringify
 {
     my ($jc, $input) = @_;
-    my $abstraction_leaked;
     if (! utf8::is_utf8 ($input)) {
+	if ($input =~ /[\x{80}-\x{FF}]/ && $jc->{_strict}) {
+	    return "Non-ASCII byte in non-utf8 string";
+	}
 	if (! valid_utf8 ($input)) {
 	    if ($jc->{_replace_bad_utf8}) {
 		# Discard the warnings from Unicode::UTF8.
@@ -145,13 +162,7 @@ sub stringify
 	    }
 	}
     }
-    $input =~ s/("|\\)/\\$1/g;
-    $input =~ s/\x08/\\b/g;
-    $input =~ s/\f/\\f/g;
-    $input =~ s/\n/\\n/g;
-    $input =~ s/\r/\\r/g;
-    $input =~ s/\t/\\t/g;
-    $input =~ s/([\x00-\x1f])/sprintf ("\\u%04x", ord ($1))/ge;
+    $input = json_escape ($input);
     if ($jc->{_escape_slash}) {
 	$input =~ s!/!\\/!g;
     }
@@ -189,10 +200,8 @@ sub call_to_json
 	return 'undefined value from user routine';
     }
     if ($jc->{_validate}) {
-#	print "Validating...\n";
 	my $error = $jc->validate_user_json ($json);
 	if ($error) {
-#	    cluck "Validating failed";
 	    return $error;
 	}
     }
@@ -200,25 +209,45 @@ sub call_to_json
     return undef;
 }
 
+# This handles a non-finite floating point number, which is either
+# nan, inf, or -inf. The return value is undefined if successful, or
+# the error value if an error occurred.
+
+sub handle_non_finite
+{
+    my ($jc, $input, $type) = @_;
+    my $handler = $jc->{_non_finite_handler};
+    if ($handler) {
+	my $output = &{$handler} ($type);
+	if (! $output) {
+	    return "Empty output from non-finite handler";
+	}
+	$jc->{output} .= $output;
+	return undef;
+    }
+    if ($jc->{_strict}) {
+	return "non-finite number";
+    }
+    $jc->{output} .= "\"$type\"";
+    return undef;
+}
+
 sub handle_number
 {
     my ($jc, $input) = @_;
-    # Perl thinks that nan, inf, etc. look like numbers,
-    # apparently.
+    # Perl thinks that nan, inf, etc. look like numbers.
     if (isnan ($input)) {
-	$jc->{output} .= '"nan"';
+	return $jc->handle_non_finite ($input, 'nan');
     }
     elsif (isinf ($input)) {
-	$jc->{output} .= '"inf"';
+	return $jc->handle_non_finite ($input, 'inf');
     }
     elsif (isneginf ($input)) {
-	$jc->{output} .= '"-inf"';
+	return $jc->handle_non_finite ($input, '-inf');
     }
     elsif (isfloat ($input)) {
-#	print "Priting floats.\n";
 	# Default format
 	if ($jc->{_fformat}) {
-#	    print "Overriing floats with $jc->{_fformat} for $input.\n";
 	    # Override. Validation is in
 	    # JSON::Create::set_fformat.
 	    $jc->{output} .= sprintf ($jc->{_fformat}, $input);
@@ -229,19 +258,46 @@ sub handle_number
     }
     else {
 	# integer or looks like integer.
-	# if ($jc->{_fformat}) {
-	#     print "Overriing integer with $jc->{_fformat} for $input.\n";
-	#     $jc->{output} .= sprintf ($jc->{_fformat}, $input);
-	# }
-	# else {
-#	    print "INT $input\n";
-	    # integer
-	    $jc->{output} .= $input;
-#	}
+	$jc->{output} .= $input;
     }
     return undef;
 }
 
+sub newline_indent
+{
+    my ($jc) = @_;
+    $jc->{output} .= "\n" . "\t" x $jc->{depth};
+}
+
+sub openB
+{
+    my ($jc, $b) = @_;
+    $jc->{output} .= $b;
+    if ($jc->{_indent}) {
+	$jc->{depth}++;
+	$jc->newline_indent ();
+    }
+}
+
+sub closeB
+{
+    my ($jc, $b) = @_;
+    if ($jc->{_indent}) {
+	$jc->{depth}--;
+	$jc->newline_indent ();
+    }
+    $jc->{output} .= $b;
+}
+
+sub comma
+{
+    my ($jc) = @_;
+    $jc->{output} .= ',';
+    if ($jc->{_indent}) {
+	$jc->newline_indent ();
+    }
+}
+
 sub create_json_recursively
 {
     my ($jc, $input) = @_;
@@ -257,10 +313,16 @@ sub create_json_recursively
     else {
 	# Break encapsulation if the user has not supplied handlers.
 	$ref = reftype ($input);
+	if ($ref && $jc->{_strict}) {
+	    my $origref = ref ($input);
+	    if ($ref ne $origref) {
+		return "Object cannot be serialized to JSON: $origref";
+	    }
+	}
     }
     if ($ref) {
 	if ($ref eq 'HASH') {
-	    $jc->{output} .= '{';
+	    $jc->openB ('{');
 	    for my $k (keys %$input) {
 		my $error = stringify ($jc, $k);
 		if ($error) {
@@ -277,12 +339,13 @@ sub create_json_recursively
 			return $error;
 		    }
 		}
-		$jc->{output} .= ',';
+		$jc->comma ();
 	    }
-	    $jc->{output} =~ s/,$/}/;
+	    $jc->{output} =~ s/,(\s*)$/$1/;
+	    $jc->closeB ('}');
 	}
 	elsif ($ref eq 'ARRAY') {
-	    $jc->{output} .= '[';
+	    $jc->openB ('[');
 	    for my $k (@$input) {
 		my $bool = isbool (\$k);
 		if ($bool) {
@@ -294,12 +357,16 @@ sub create_json_recursively
 			return $error;
 		    }
 		}
-		$jc->{output} .= ',';
+		$jc->comma ();
 	    }
-	    $jc->{output} =~ s/,$/]/;
+	    $jc->{output} =~ s/,(\s*)$/$1/;
+	    $jc->closeB (']');
 	}
 	elsif ($ref eq 'SCALAR') {
-	    $error = stringify ($jc, $$input);
+	    if ($jc->{_strict}) {
+		return "Input's type cannot be serialized to JSON";
+	    }
+	    $error = $jc->create_json_recursively ($$input);
 	    if ($error) {
 		return $error;
 	    }
@@ -314,7 +381,6 @@ sub create_json_recursively
 		}
 		else {
 		    my $handler = $jc->{_handlers}{$ref};
-#		    printf ("%s:%d: %s\n", __FILE__, __LINE__, $handler);
 		    if ($handler) {
 			if ($handler eq 'bool') {
 			    if ($$input) {
@@ -354,7 +420,7 @@ sub create_json_recursively
     }
     else {
 	my $error;
-	if (looks_like_number ($input) && $input !~ /^0/) {
+	if (looks_like_number ($input) && $input !~ /^0[^.]/) {
 	    $error = $jc->handle_number ($input);
 	}
 	else {
@@ -381,10 +447,10 @@ sub user_error
 sub create_json
 {
     my ($input, %options) = @_;
-    my $output;
     my $jc = bless {
 	output => '',
     };
+    $jc->{_strict} = !! $options{strict};
     my $error = create_json_recursively ($jc, $input);
     if ($error) {
 	$jc->user_error ($error);
@@ -394,6 +460,13 @@ sub create_json
     return $jc->{output};
 }
 
+sub create_json_strict
+{
+    my ($input, %options) = @_;
+    $options{strict} = 1;
+    return create_json ($input, %options);
+}
+
 sub new
 {
     return bless {
@@ -401,18 +474,30 @@ sub new
     };
 }
 
+sub strict
+{
+    my ($jc, $onoff) = @_;
+    $jc->{_strict} = !! $onoff;
+}
+
 sub get_handlers
 {
     my ($jc) = @_;
     return $jc->{_handlers};
 }
 
+sub non_finite_handler
+{
+    my ($jc, $handler) = @_;
+    $jc->{_non_finite_handler} = $handler;
+    return undef;
+}
+
 sub obj
 {
     my ($jc, %things) = @_;
     my $handlers = $jc->get_handlers ();
     for my $k (keys %things) {
-#	printf "%s:%d: Adding key %s for %s.\n", __FILE__, __LINE__, $things{$k}, $k;
 	$handlers->{$k} = $things{$k};
     }
 }
@@ -434,8 +519,6 @@ sub escape_slash
 
 sub set_fformat_unsafe
 {
-#    printf ("%s:%d: ", __FILE__, __LINE__);
-#    print ("@_ OK\n");
     my ($jc, $fformat) = @_;
     if ($fformat) {
 	$jc->{_fformat} = $fformat;
@@ -473,7 +556,6 @@ sub type_handler
 sub obj_handler
 {
     my ($jc, $handler) = @_;
-#    print "Handler for object set.\n";
     $jc->{_obj_handler} = $handler;
 }
 
@@ -518,4 +600,10 @@ sub validate
     return JSON::Create::validate (@_);
 }
 
+sub indent
+{
+    my ($jc, $onoff) = @_;
+    $jc->{_indent} = !! $onoff;
+}
+
 1;
